import numpy as np
from pystencils import Field, Assignment, AssignmentCollection
from pystencils.slicing import normalize_slice, get_periodic_boundary_src_dst_slices

from pystencils.gpucuda import make_python_function as make_python_function_gpu
from pystencils.cpu import make_python_function as make_python_function_cpu
from pystencils import create_kernel
from lbmpy.methods.abstractlbmethod import LbmCollisionRule


def create_copy_kernel(domain_size, from_slice, to_slice, stencil, index_dimensions=0, index_dim_shape=1,
                       dtype=np.float64, target=None, update_rule_assignments=None, reads_writes="pull"):
    """Copies a rectangular part of an array to another non-overlapping part"""
    if index_dimensions not in (0, 1):
        raise NotImplementedError("Works only for one or zero index coordinates")

    f = Field.create_generic("pdfs", len(domain_size), index_dimensions=index_dimensions, dtype=dtype)
    normalized_from_slice = normalize_slice(from_slice, f.spatial_shape)
    normalized_to_slice = normalize_slice(to_slice, f.spatial_shape)

    offset = [s1.start - s2.start for s1, s2 in zip(normalized_from_slice, normalized_to_slice)]
    assert offset == [s1.stop - s2.stop for s1, s2 in zip(normalized_from_slice, normalized_to_slice)], \
        "Slices have to have same size"

    update_eqs = set()
    if update_rule_assignments is None:
        for i in range(index_dim_shape):
            eq = Assignment(f(i), f[tuple(offset)](i))
            update_eqs.add(eq)
    else:
        for update_rule_assignment in update_rule_assignments:
            if reads_writes == "pull":
                field_accesses = update_rule_assignment.rhs.atoms(Field.Access)
            else:
                field_accesses = update_rule_assignment.lhs.atoms(Field.Access)

            for field_access in field_accesses:
                if 0 in stencil:
                    for stencil_part, off in zip(stencil, field_access.offsets):
                        if (stencil_part < 0 and off < 0) or (stencil_part > 0 and off > 0):
                            if reads_writes == "pull":
                                eq = Assignment(f(field_access.index[0]), f[tuple(offset)](field_access.index[0]))
                                update_eqs.add(eq)
                            if reads_writes == "push":
                                eq = Assignment(f[tuple(offset)](field_access.index[0]), f(field_access.index[0]))
                                update_eqs.add(eq)
                else:
                    if stencil == field_access.offsets:
                        if reads_writes == "pull":
                            eq = Assignment(f(field_access.index[0]), f[tuple(offset)](field_access.index[0]))
                            update_eqs.add(eq)
                        if reads_writes == "push":
                            eq = Assignment(f[tuple(offset)](field_access.index[0]), f(field_access.index[0]))
                            update_eqs.add(eq)

    kernels = []
    if len(update_eqs) > 0:
        ast = create_kernel(list(update_eqs), iteration_slice=to_slice,
                            skip_independence_check=True, target=target)
        if target == 'cpu':
            kernels.append(make_python_function_cpu(ast))
        else:
            kernels.append(make_python_function_gpu(ast))

    return kernels


def get_periodic_boundary_functor(stencil, domain_size, index_dimensions=0, index_dim_shape=1, ghost_layers=1,
                                  thickness=None, dtype=float, target=None, update_rule=None, reads_writes="pull"):

    # sort stencil entry's by number of zeros to make sure, that surfaces, edges and corners are
    # communicated in that order. Otherwise values (e.g. corner values) will be overwritten
    stencil.sort(key=number_of_zeros, reverse=True)

    src_dst_slice_tuples = get_periodic_boundary_src_dst_slices(stencil, ghost_layers, thickness)
    kernels = []

    if isinstance(update_rule, Assignment):
        update_rule_assignments = [update_rule]
    elif isinstance(update_rule, (AssignmentCollection, LbmCollisionRule)):
        update_rule_assignments = update_rule.all_assignments
    else:
        update_rule_assignments = None

    # by default we only get the communication pattern for pull stream
    if reads_writes is not "pull" or reads_writes is not "push":
        assert "not a valid argument for the definition of read or/and write access"

    for i, (src_slice, dst_slice) in enumerate(src_dst_slice_tuples):
        for j in create_copy_kernel(domain_size, src_slice, dst_slice, stencil[i], index_dimensions,
                                    index_dim_shape, dtype, target, update_rule_assignments, reads_writes):
            kernels.append(j)

    def functor(pdfs, **_):
        for kernel in kernels:
            kernel(pdfs=pdfs)

    return functor


def number_of_zeros(stencil_entry):
    count_zeros = 0
    for tmp in stencil_entry:
        if tmp == 0:
            count_zeros += 1

    return count_zeros
